package models

import java.util.Date
import play.api.Play.current
import play.api.db.slick.Config.driver.simple._
import slick.lifted.{ Join, MappedTypeMapper }
import scala.slick.lifted.ForeignKeyAction

case class Page[A](items: Seq[A], page: Int, offset: Long, total: Long) {
  lazy val prev = Option(page - 1).filter(_ >= 0)
  lazy val next = Option(page + 1).filter(_ => (offset + items.size) < total)
}

/**
 * Data Access Object trait
 *
 *  Used to create the DAOs: Companies and Computers
 */
private[models] trait DAO extends ComputersComponent with CompaniesComponent {
  val Companies = new Companies 
  val Computers = new Computers 
}

case class Company(id: Option[Long], name: String)

case class Computer(id: Option[Long] = None, name: String, introduced: Option[Date] = None, discontinued: Option[Date] = None, companyId: Option[Long] = None)

trait CompaniesComponent {
  val Companies: Companies
  
  class Companies extends Table[Company]("COMPANY") {
    def id = column[Long]("id", O.PrimaryKey, O.AutoInc)
    def name = column[String]("name", O.NotNull)
    def * = id.? ~ name <> (Company.apply _, Company.unapply _)
    def autoInc = * returning id
  }
}

object Companies extends DAO {
  /**
   * Construct the Map[String,String] needed to fill a select options set
   */
  def options(implicit s: Session): Seq[(String, String)] = {
    val query = (for {
      company <- Companies
    } yield (company.id, company.name)).sortBy(_._2)
    query.list.map(row => (row._1.toString, row._2))
  }

  /**
   * Insert a new company
   * @param company
   */
  def insert(company: Company)(implicit s: Session) {
    Companies.autoInc.insert(company)
  }
}

trait ComputersComponent { self: CompaniesComponent =>
  val Computers: Computers

  class Computers extends Table[Computer]("COMPUTER") {
    implicit val javaUtilDateTypeMapper = MappedTypeMapper.base[java.util.Date, java.sql.Date](
      x => new java.sql.Date(x.getTime),
      x => new java.util.Date(x.getTime))

    def id = column[Long]("id", O.PrimaryKey, O.AutoInc)
    def name = column[String]("name", O.NotNull)
    def introduced = column[Date]("introduced", O.Nullable)
    def discontinued = column[Date]("discontinued", O.Nullable)
    def companyId = column[Long]("companyId", O.Nullable)
    def companyFk = foreignKey("company_id", id, Companies)(_.id, onDelete = ForeignKeyAction.Cascade)

    def * = id.? ~ name ~ introduced.? ~ discontinued.? ~ companyId.? <> (Computer.apply _, Computer.unapply _)

    def autoInc = * returning id

    val byId = createFinderBy(_.id)
  }
}

object Computers extends DAO {
  /**
   * Retrieve a computer from the id
   * @param id
   */
  def findById(id: Long)(implicit s: Session): Option[Computer] =
    Computers.byId(id).firstOption

  /**
   * Count all computers
   */
  def count(implicit s: Session): Int =
    Query(Computers.length).first

  /**
   * Count computers with a filter
   * @param filter
   */
  def count(filter: String)(implicit s: Session): Int =
    Query(Computers.where(_.name.toLowerCase like filter.toLowerCase).length).first

  /**
   * Return a page of (Computer,Company)
   * @param page
   * @param pageSize
   * @param orderBy
   * @param filter
   */
  def list(page: Int = 0, pageSize: Int = 10, orderBy: Int = 1, filter: String = "%")(implicit s: Session): Page[(Computer, Option[Company])] = {

    val offset = pageSize * page
    val query =
      (for {
        (computer, company) <- Computers leftJoin Companies on (_.companyId === _.id)
        if computer.name.toLowerCase like filter.toLowerCase()
      } yield (computer, company.id.?, company.name.?))
        .drop(offset)
        .take(pageSize)

    val totalRows = count(filter)
    val result = query.list.map(row => (row._1, row._2.map(value => Company(Option(value), row._3.get))))

    Page(result, page, offset, totalRows)
  }

  /**
   * Insert a new computer
   * @param computer
   */
  def insert(computer: Computer)(implicit s: Session) {
    Computers.autoInc.insert(computer)
  }

  /**
   * Update a computer
   * @param id
   * @param computer
   */
  def update(id: Long, computer: Computer)(implicit s: Session) {
    val computerToUpdate: Computer = computer.copy(Some(id))
    Computers.where(_.id === id).update(computerToUpdate)
  }

  /**
   * Delete a computer
   * @param id
   */
  def delete(id: Long)(implicit s: Session) {
    Computers.where(_.id === id).delete
  }
}



